use std::collections::HashMap;
use std::env;
use std::fs;
use rand::Rng;

type Word = String;
type NextWords = Vec<(Word, usize)>;

struct MarkovChain { transitions: HashMap<Word, NextWords> }

impl MarkovChain {
    fn new() -> Self {
        MarkovChain { transitions: HashMap::new() }
    }

    fn add_transition(&mut self, word: &str, next_word: &str) {
        let entry = self.transitions.entry(word.to_string())
            .or_insert_with(Vec::new);
        if
            let Some((_, count)) = entry.iter_mut()
                .find(|(w, _)| w == next_word)
        {
            *count += 1;
        } else { entry.push((next_word.to_string(), 1)) }
    }

    fn choose_next_word(&self, word: &str) -> Option<&str> {
        let next_words = self.transitions.get(word)?;

        let total_count: usize = next_words.iter()
            .map(|(_, count)| count)
            .sum();

        let mut rng = rand::thread_rng();
        let mut choice = rng.gen_range(0..total_count);

        for (next_word, count) in next_words {
            if choice < *count { return Some(next_word); }
            choice -= count;
        }

        return None
    }

    fn build_chain(text: &str) -> Self {
        let mut chain = MarkovChain::new();

        let mut words = Vec::new();
        for line in text.lines() {
            for word in line.split_whitespace() {
                words.push(word.to_string());
            }
            words.push("\n".to_string());
        }

        for window in words.windows(2) {
            if let [word, next_word] = window {
                chain.add_transition(word, next_word);
            }
        }

        return chain
    }

    fn generate_text(&self, start_word: &str, length: usize) -> String {
        let mut result = Vec::new();
        let mut current_word = start_word;

        for _ in 0..length {
            if current_word == "\n" {
                result.push("\n".to_string());
            } else {
                result.push(format!("{} ", current_word));
            }

            if let Some(next_word) = self.choose_next_word(current_word) {
                current_word = next_word;
            } else { break; }
        }

        return result.join("").replace(" \n", "\n") // Remove space at end of line.
    }
}

fn main() {
    let args: Vec<String> = env::args().collect();
    if args.len() < 4 {
        eprintln!("Usage:\n{} <file> <start word> <length>", args[0]);
        return;
    }

    let input_file = &args[1];
    let start_word = &args[2];
    let length: usize = args[3].parse().expect("Length must be number.");

    let text = fs::read_to_string(input_file).expect("Could not read input file.");

    let chain = MarkovChain::build_chain(&text);

    let generated_text = chain.generate_text(start_word, length);
    println!("{}", generated_text);
}

